use errors::*;

use std::io;
use std::process::Command;
use clap::ArgMatches;
use path::{self, FilePath};
use config::{Config, FileOptions};
use fs;

pub fn run() -> Result<()> {
    let matches = clap_app!(koi =>
        (about: crate_description!())
        (version: crate_version!())
        (author: crate_authors!())
        (@setting AllowExternalSubcommands)
        (@setting ColorNever)
        (@setting DeriveDisplayOrder)
        (@setting GlobalVersion)
        (@subcommand new =>
            (about: "Create a new repository")
        )
        (@subcommand add =>
            (about: "Add files to repository")
            (@arg files: +required ... "Files to add")
        )
        (@subcommand remove =>
            (about: "Remove files from repository")
            (@arg files: +required ... "Files to remove")
        )
        (@subcommand link =>
            (about: "Link files to home directory")
            (@arg force: -f --force "Remove existing files")
        )
        (@subcommand list =>
            (about: "List files in repository")
        )
    ).get_matches();

    match matches.subcommand() {
        ("new", Some(m)) => run_new(m),
        ("add", Some(m)) => run_add(m),
        ("remove", Some(m)) => run_remove(m),
        ("link", Some(m)) => run_link(m),
        ("list", Some(m)) => run_list(m),
        (ext, Some(m)) => run_external(ext, m),
        _ => unreachable!(),
    }
}

fn run_new(_matches: &ArgMatches) -> Result<()> {
    fs::ensure_tree(&path::repo_file()?)?;
    Ok(Config::new().write()?)
}

fn run_add(matches: &ArgMatches) -> Result<()> {
    let mut cfg = Config::read()?;

    let files = values_t!(matches, "files", String)?;
    for file in &files {
        let loc = FilePath::locate(file)?;
        if cfg.files.has(&loc.name) {
            eprintln!("File found in repository");
            continue;
        }

        if !loc.home.is_file() {
            eprintln!("Not a file");
            continue;
        }

        fs::ensure_tree(&loc.repo)?;
        fs::rename(false, &loc.home, &loc.repo)?;
        fs::symlink(false, &loc.repo, &loc.home)?;

        cfg.files.add(&loc.name, FileOptions { ignore: false });
    }

    Ok(cfg.write()?)
}

fn run_remove(matches: &ArgMatches) -> Result<()> {
    let mut cfg = Config::read()?;

    let files = values_t!(matches, "files", String)?;
    for file in &files {
        let loc = FilePath::locate(file)?;
        if !cfg.files.has(&loc.name) {
            eprintln!("File not found in repository");
            continue;
        }

        if !loc.repo.is_file() {
            eprintln!("Not a file");
            continue;
        }

        fs::rename(false, &loc.repo, &loc.home)?;

        cfg.files.remove(&loc.name);
    }

    Ok(cfg.write()?)
}

fn run_link(matches: &ArgMatches) -> Result<()> {
    let cfg = Config::read()?;

    for (file, options) in cfg.files.iter() {
        let loc = FilePath::locate_buf(file)?;
        if options.ignore {
            println!("Skipping ignored file");
            continue;
        }

        fs::ensure_tree(&loc.home)?;
        fs::symlink(value_t!(matches, "force", bool)?, &loc.repo, &loc.home)?;
    }

    Ok(())
}

fn run_list(_matches: &ArgMatches) -> Result<()> {
    let cfg = Config::read()?;

    for (file, options) in cfg.files.iter() {
        println!("- {:?}\n  Ignore: {}", file, options.ignore);
    }

    Ok(())
}

fn run_external(ext: &str, matches: &ArgMatches) -> Result<()> {
    let cmd = "koi-".to_owned() + ext;
    let args = match values_t!(matches, "", String) {
        Ok(v) => v,
        Err(_) => vec![String::from("--help")],
    };

    match Command::new(&cmd)
        .args(args)
        .current_dir(path::repo_dir()?)
        .spawn() {
        Ok(_) => Ok(()),
        Err(e) => {
            if e.kind() == io::ErrorKind::NotFound {
                Err(
                    format!("failed to run external command: command {} not found", &cmd).into(),
                )
            } else {
                Err(format!("failed to run external command: {}", e).into())
            }
        }
    }
}
